@@ -27,15 +27,16 @@ static char *default_backend = NULL;
 #define at_least_one(a) ({ __typeof__(a) __at_least_one=(a); (__at_least_one)?__at_least_one:1; })
 
 static void logger(struct crypt_options *options, int class, char *format, ...) {
-        va_list argp;
-        char *target;
+	va_list argp;
+	char *target = NULL;
 
-        va_start(argp, format);
-        vasprintf(&target, format, argp);
-        options->icb->log(class, target);
+	va_start(argp, format);
 
-        va_end(argp);
-        free(target);
+	if (vasprintf(&target, format, argp) > 0)
+		options->icb->log(class, target);
+
+	va_end(argp);
+	free(target);
 }
 
 static void hexprintICB(struct crypt_options *options, int class, char *d, int n)
@@ -107,7 +108,7 @@ static char *process_key(struct crypt_options *options, char *pass, int passLen)
 		memcpy(key,pass,options->key_size);
 		return key;
 	}
-	
+
 	/* key is coming from tty, fd or binary stdin */
 	if (options->hash) {
 		if (hash(NULL, options->hash,
@@ -198,6 +199,34 @@ out:
 	return ret;
 }
 
+static int wipe_device_header(const char *device, int sectors)
+{
+	char *buffer;
+	int size = sectors * SECTOR_SIZE;
+	int r = -1;
+	int devfd;
+
+	devfd = open(device, O_RDWR | O_DIRECT | O_SYNC);
+	if(devfd == -1) {
+		set_error("Can't wipe header on device %s", device);
+		return -EINVAL;
+	}
+
+	buffer = malloc(size);
+	if (!buffer) {
+		close(devfd);
+		return -ENOMEM;
+	}
+	memset(buffer, 0, size);
+
+	r = write_blockwise(devfd, buffer, size) < size ? -EIO : 0;
+
+	free(buffer);
+	close(devfd);
+
+	return r;
+}
+
 static int parse_into_name_and_mode(const char *nameAndMode, char *name,
 				    char *mode)
 {
@@ -226,6 +255,33 @@ static int parse_into_name_and_mode(const char *nameAndMode, char *name,
 #undef str
 #undef xstr
 }
+
+/* Select free keyslot or verifies that the one specified is empty */
+static int keyslot_from_option(int keySlotOption, struct luks_phdr *hdr, struct crypt_options *options) {
+        if(keySlotOption >= 0) {
+                if(keySlotOption >= LUKS_NUMKEYS) {
+                        logger(options,CRYPT_LOG_ERROR,"slot %d too high, please pick between 0 and %d", keySlotOption, LUKS_NUMKEYS);
+                        return -EINVAL;
+                } else if(hdr->keyblock[keySlotOption].active != LUKS_KEY_DISABLED) {
+                        logger(options,CRYPT_LOG_ERROR,"slot %d full, please pick another one", keySlotOption);
+                        return -EINVAL;
+                } else {
+                        return keySlotOption;
+                }
+        } else {
+                int i;
+                /* Find empty key slot */
+                for(i=0; i<LUKS_NUMKEYS; i++) {
+                        if(hdr->keyblock[i].active == LUKS_KEY_DISABLED) break;
+                }
+                if(i==LUKS_NUMKEYS) {
+                        logger(options,CRYPT_LOG_ERROR,"All slots full");
+                        return -EINVAL;
+                }
+                return i;
+        }
+}
+
 static int __crypt_create_device(int reload, struct setup_backend *backend,
                                  struct crypt_options *options)
 {
@@ -234,7 +290,7 @@ static int __crypt_create_device(int reload, struct setup_backend *backend,
 	};
 	struct device_infos infos;
 	char *key = NULL;
-	int keyLen;
+	unsigned int keyLen;
 	char *processed_key = NULL;
 	int r;
 
@@ -244,7 +300,7 @@ static int __crypt_create_device(int reload, struct setup_backend *backend,
 			return r;
 	} else {
 		if (r >= 0) {
-			set_error("Device already exists");
+			set_error("Device %s already exists.", options->name);
 			return -EEXIST;
 		}
 		if (r != -ENODEV)
@@ -280,24 +336,24 @@ static int __crypt_create_device(int reload, struct setup_backend *backend,
 		set_error("Key reading error");
 		return -ENOENT;
 	}
-	
+
 	processed_key = process_key(options,key,keyLen);
 	safe_free(key);
-	
+
 	if (!processed_key) {
 		const char *error=get_error();
 		if(error) {
-			char *c_error_handling_sucks;
-			asprintf(&c_error_handling_sucks,"Key processing error: %s",error);
-			set_error(c_error_handling_sucks);
+			char *c_error_handling_sucks = NULL;
+			if (asprintf(&c_error_handling_sucks,"Key processing error: %s",error) > 0)
+				set_error(c_error_handling_sucks);
 			free(c_error_handling_sucks);
 		} else
 			set_error("Key processing error");
 		return -ENOENT;
 	}
-	
-	r = backend->create(reload, options, processed_key);
-	
+
+	r = backend->create(reload, options, processed_key, NULL);
+
 	safe_free(processed_key);
 
 	return r;
@@ -349,7 +405,7 @@ static int __crypt_resize_device(int details, struct setup_backend *backend,
 	if (infos.readonly)
 		options->flags |= CRYPT_FLAG_READONLY;
 
-	r = backend->create(1, &tmp, key);
+	r = backend->create(1, &tmp, key, NULL);
 
 	safe_free(key);
 
@@ -369,28 +425,27 @@ static int __crypt_remove_device(int arg, struct setup_backend *backend,
 		return -EBUSY;
 	}
 
-	return backend->remove(options);
+	return backend->remove(0, options);
 }
 
 static int __crypt_luks_format(int arg, struct setup_backend *backend, struct crypt_options *options)
 {
 	int r;
-	
+
 	struct luks_phdr header;
 	struct luks_masterkey *mk=NULL;
 	char *password=NULL; 
 	char cipherName[LUKS_CIPHERNAME_L];
 	char cipherMode[LUKS_CIPHERMODE_L];
-	int passwordLen;
+	unsigned int passwordLen;
 	int PBKDF2perSecond;
-	
-	if (!LUKS_device_ready(options->device, O_RDWR | O_EXCL)) {
-		set_error("Can not access device");
-		r = -ENOTBLK; goto out;
-	}
+        int keyIndex;
+
+	if (!LUKS_device_ready(options->device, O_RDWR | O_EXCL))
+		return -ENOTBLK;
 
 	mk = LUKS_generate_masterkey(options->key_size);
-	if(NULL == mk) return -ENOMEM; 
+	if(NULL == mk) return -ENOMEM; // FIXME This may be misleading, since we don't know what went wrong
 
 #ifdef LUKS_DEBUG
 #define printoffset(entry) logger(options, CRYPT_LOG_ERROR, ("offset of " #entry " = %d\n", (char *)(&header.entry)-(char *)(&header))
@@ -419,8 +474,13 @@ static int __crypt_luks_format(int arg, struct setup_backend *backend, struct cr
 		return r; 
 	}
 
+        keyIndex = keyslot_from_option(options->key_slot, &header, options);
+        if(keyIndex == -EINVAL) {
+                r = -EINVAL; goto out;
+        }
+
 	PBKDF2perSecond = LUKS_benchmarkt_iterations();
-	header.keyblock[0].passwordIterations = at_least_one(PBKDF2perSecond * ((float)options->iteration_time / 1000.0));
+	header.keyblock[keyIndex].passwordIterations = at_least_one(PBKDF2perSecond * ((float)options->iteration_time / 1000.0));
 #ifdef LUKS_DEBUG
 	logger(options->icb->log,CRYPT_LOG_ERROR, "pitr %d\n", header.keyblock[0].passwordIterations);
 #endif
@@ -429,8 +489,12 @@ static int __crypt_luks_format(int arg, struct setup_backend *backend, struct cr
 		r = -EINVAL; goto out;
 	}
 
+	/* Wipe first 8 sectors - fs magic numbers etc. */
+	r = wipe_device_header(options->device, 8);
+	if(r < 0) goto out;
+
 	/* Set key, also writes phdr */
-	r = LUKS_set_key(options->device, options->key_slot==-1?0:(unsigned int)options->key_slot, password, passwordLen, &header, mk, backend);
+	r = LUKS_set_key(options->device, keyIndex, password, passwordLen, &header, mk, backend);
 	if(r < 0) goto out; 
 
 	r = 0;
@@ -444,24 +508,25 @@ static int __crypt_luks_open(int arg, struct setup_backend *backend, struct cryp
 {
 	struct luks_masterkey *mk=NULL;
 	struct luks_phdr hdr;
-	char *password; int passwordLen;
+	char *prompt = NULL;
+	char *password;
+	unsigned int passwordLen;
 	struct device_infos infos;
 	struct crypt_options tmp = {
 		.name = options->name,
 	};
-	char *dmCipherSpec;
+	char *dmCipherSpec = NULL;
 	int r, tries = options->tries;
-	
+	int excl = (options->flags & CRYPT_FLAG_NON_EXCLUSIVE_ACCESS) ? 0 : O_EXCL ;
+
 	r = backend->status(0, &tmp, NULL);
 	if (r >= 0) {
-		set_error("Device already exists");
+		set_error("Device %s already exists.", options->name);
 		return -EEXIST;
 	}
 
-	if (!LUKS_device_ready(options->device, O_RDONLY | O_EXCL)) {
-		set_error("Can not access device");
+	if (!LUKS_device_ready(options->device, O_RDONLY | excl))
 		return -ENOTBLK;
-	}
 
 	if (get_device_infos(options->device, &infos) < 0) {
 		set_error("Can't get device information.\n");
@@ -471,10 +536,13 @@ static int __crypt_luks_open(int arg, struct setup_backend *backend, struct cryp
 	if (infos.readonly)
 		options->flags |= CRYPT_FLAG_READONLY;
 
+	if(asprintf(&prompt, "Enter LUKS passphrase for %s: ", options->device) < 0)
+		return -ENOMEM;
+
 start:
 	mk=NULL;
 
-	if(get_key("Enter LUKS passphrase: ",&password,&passwordLen, 0, options->key_file,  options->passphrase_fd, options->timeout, options->flags))
+	if(get_key(prompt, &password, &passwordLen, 0, options->key_file, options->passphrase_fd, options->timeout, options->flags))
 		tries--;
 	else
 		tries = 0;
@@ -482,18 +550,18 @@ start:
 	if(!password) {
 		r = -EINVAL; goto out;
 	}
-        
-        r = LUKS_open_any_key(options->device, password, passwordLen, &hdr, &mk, backend);
-	if(r < 0) {
-		set_error("No key available with this passphrase.\n");
-		goto out1;
-	} else
-                logger(options, CRYPT_LOG_NORMAL,"key slot %d unlocked.\n", r);
 
-	
+        r = LUKS_open_any_key(options->device, password, passwordLen, &hdr, &mk, backend);
+	if (r == -EPERM)
+		set_error("No key available with this passphrase.\n");
+	if (r < 0)
+		goto out1;
+
+	logger(options, CRYPT_LOG_NORMAL,"key slot %d unlocked.\n", r);
+
+
 	options->offset = hdr.payloadOffset;
- 	asprintf(&dmCipherSpec, "%s-%s", hdr.cipherName, hdr.cipherMode);
-	if(!dmCipherSpec) {
+ 	if (asprintf(&dmCipherSpec, "%s-%s", hdr.cipherName, hdr.cipherMode) < 0) {
 		r = -ENOMEM;
 		goto out2;
 	}
@@ -511,10 +579,13 @@ start:
 		r = -EINVAL; goto out2;
 	}
 	options->size -= options->offset;
-	r = backend->create(0, options, mk->key);
+	/* FIXME: code allows multiple crypt mapping, cannot use uuid then.
+	 * anyway, it is dangerous and can corrupt data. Remove it in next version! */
+	r = backend->create(0, options, mk->key, excl ? hdr.uuid : NULL);
 
  out2:
 	free(dmCipherSpec);
+	dmCipherSpec = NULL;
  out1:
 	safe_free(password);
  out:
@@ -522,6 +593,8 @@ start:
 	if (r == -EPERM && tries > 0)
 		goto start;
 
+	free(prompt);
+
 	return r;
 }
 
@@ -530,39 +603,20 @@ static int __crypt_luks_add_key(int arg, struct setup_backend *backend, struct c
 	struct luks_masterkey *mk=NULL;
 	struct luks_phdr hdr;
 	char *password=NULL; unsigned int passwordLen;
-	unsigned int i; unsigned int keyIndex;
+        unsigned int keyIndex;
 	const char *device = options->device;
 	int r;
-	int key_slot = options->key_slot;
-	
-	if (!LUKS_device_ready(options->device, O_RDWR)) {
-		set_error("Can not access device");
-		r = -ENOTBLK; goto out;
-	}
+
+	if (!LUKS_device_ready(options->device, O_RDWR))
+		return -ENOTBLK;
 
 	r = LUKS_read_phdr(device, &hdr);
 	if(r < 0) return r;
 
-        if(key_slot != -1) {
-                if(key_slot >= LUKS_NUMKEYS) {
-                        set_error("slot %d too high, please pick between 0 and %d", key_slot, LUKS_NUMKEYS);
-                        return -EINVAL;
-                } else if(hdr.keyblock[key_slot].active != LUKS_KEY_DISABLED) {
-                        set_error("slot %d full, please pick another one", key_slot);
-                        return -EINVAL;
-                } else {
-                        keyIndex = key_slot;
-                }
-        } else {
-                /* Find empty key slot */
-                for(i=0; i<LUKS_NUMKEYS; i++) {
-                        if(hdr.keyblock[i].active == LUKS_KEY_DISABLED) break;
-                }
-                if(i==LUKS_NUMKEYS) {
-                        set_error("All slots full");
-                        return -EINVAL;
-                }
-                keyIndex = i;
+
+        keyIndex = keyslot_from_option(options->key_slot, &hdr, options);
+        if(keyIndex == -EINVAL) {
+                r = -EINVAL; goto out;
         }
 
 	get_key("Enter any LUKS passphrase: ",
@@ -585,7 +639,7 @@ static int __crypt_luks_add_key(int arg, struct setup_backend *backend, struct c
 	        logger(options, CRYPT_LOG_NORMAL,"key slot %d unlocked.\n", r);
 
 	safe_free(password);
-	
+
 	get_key("Enter new passphrase for key slot: ",
                 &password,
                 &passwordLen,
@@ -619,11 +673,10 @@ static int luks_remove_helper(int arg, struct setup_backend *backend, struct cry
 	const char *device = options->device;
 	int keyIndex;
 	int openedIndex;
-	int r;
-	if (!LUKS_device_ready(options->device, O_RDWR)) {
-	    set_error("Can not access device");
-	    r = -ENOTBLK; goto out;
-	}
+	int r, last_slot;
+
+	if (!LUKS_device_ready(options->device, O_RDWR))
+	    return -ENOTBLK;
 
 	if(supply_it) {
 	    get_key("Enter LUKS passphrase to be deleted: ",&password,&passwordLen, 0, options->new_key_file, options->passphrase_fd, options->timeout, options->flags);
@@ -641,11 +694,10 @@ static int luks_remove_helper(int arg, struct setup_backend *backend, struct cry
 	    keyIndex = options->key_slot;
 	}
 
-	if(LUKS_is_last_keyslot(options->device, keyIndex) && 
-	   !(options->icb->yesDialog(_("This is the last keyslot. Device will become unusable after purging this key.")))) {
-		r = -EINVAL;
-		goto out;
-	} 
+	last_slot = LUKS_is_last_keyslot(options->device, keyIndex);
+	if(last_slot && !(options->icb->yesDialog(_("This is the last keyslot. Device will become unusable after purging this key.")))) {
+		r = -EINVAL; goto out;
+	}
 
 	if(options->flags & CRYPT_FLAG_VERIFY_ON_DELKEY) {
 		options->flags &= ~CRYPT_FLAG_VERIFY_ON_DELKEY;
@@ -653,17 +705,27 @@ static int luks_remove_helper(int arg, struct setup_backend *backend, struct cry
 		if(!password) {
 			r = -EINVAL; goto out;
 		}
-		openedIndex = LUKS_open_any_key(device, password, passwordLen, &hdr, &mk, backend);
+
+                r = LUKS_read_phdr(device, &hdr);
+                if(r < 0) { 
+                        options->icb->log(CRYPT_LOG_ERROR,"Failed to access device.\n");
+                        r = -EIO; goto out;
+                }
+
+		if(!last_slot)
+			hdr.keyblock[keyIndex].active = LUKS_KEY_DISABLED;
+
+		openedIndex = LUKS_open_any_key_with_hdr(device, password, passwordLen, &hdr, &mk, backend);
                 /* Clean up */
                 if (openedIndex >= 0) {
                         LUKS_dealloc_masterkey(mk);
                         mk = NULL;
                 }
-		if(openedIndex < 0 || keyIndex == openedIndex) {
+		if(openedIndex < 0) {
                             options->icb->log(CRYPT_LOG_ERROR,"No remaining key available with this passphrase.\n");
 			    r = -EPERM; goto out;
 		} else
-                        logger(options, CRYPT_LOG_NORMAL,"key slot %d verified.\n", keyIndex);
+                        logger(options, CRYPT_LOG_NORMAL,"key slot %d verified.\n", openedIndex);
 	}
 	r = LUKS_del_key(device, keyIndex);
 	if(r < 0) goto out;
@@ -692,7 +754,10 @@ static int crypt_job(int (*job)(int arg, struct setup_backend *backend,
 
 	backend = get_setup_backend(default_backend);
 
-	setup_enter(backend,options->icb->log);
+	if (setup_enter(backend,options->icb->log) < 0) {
+		r = -ENOSYS;
+		goto out;
+	}
 
 	if (!backend) {
 		set_error("No setup backend available");
@@ -819,7 +884,7 @@ int crypt_luksDump(struct crypt_options *options)
 
 			logger(options, CRYPT_LOG_NORMAL, "\tKey material offset:\t%d\n",hdr.keyblock[i].keyMaterialOffset);
 			logger(options, CRYPT_LOG_NORMAL, "\tAF stripes:            \t%d\n",hdr.keyblock[i].stripes);
-		}		
+		}
 		else 
 			logger(options, CRYPT_LOG_NORMAL, "Key Slot %d: DISABLED\n",i);
 	}
